/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *  C.hpp - demonstration of canonical class structure       *
 *                                                           *
 *  Processing:                                              *
 *  - class defines objects with names but doesn't do any-   *
 *    thing very useful.                                     *
 *  - created to illustrate minimum functionality needed to  *
 *    build a class which actes like a type                  *
 *                                                           *
 *  Author: Jim Fawcett                                      *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#ifndef C_HPP            // don't declare multiple times
#define C_HPP
#include <iostream.h>    // cin, cout, cerr
#include <stdlib.h>      // exit
#include <string.h>      // strlen, strcat, strcpy, strncpy


typedef char* T;         // used to make meta type compilable

class C {

  private:               // encapsulated elements

    char *s;             // data members should usually be private
                         // use protected if C may be a base class
                         // for derivation

  public:                // public class interface

  // constructors - must be member function with no return type
    C(void);             // C c[10]   - define arrays of objects
    C(C& c);             // C c2 = c1 - copy constructor
    C(T& t);             // C c(t)    - promotes object t of
                         //             type T to a C object

  // destructor - must be member function taking void with no
  //              return type
    ~C(void);

  // unary operator defined as a member of C
    C& operator= (C& c); // c3 = c3 = c1
                         // returning reference allows chaining
                         // equates

  // relational friend operator - will be inline because
  // definition is placed here in header file
    friend int operator==(C& c1, C& c2)
                             { return strcmp(c1.s,c2.s) == 0; }
/**/
  // binary operator defined as a member of C
    C operator-(C& c);
                         // c2 - c1 ==> c2.operator-(c1)
                         // adds c1 to self and returns copy
                         // so first argument must be C object
                         // second argument can be "promotable"
                         // to a C object (third constructor)

  // binary operator defined as a friend of C
    friend C operator+(C& c1, C& c2);
                         // c1 + c2 ==> operator+(c1, c2)
                         // adds c1 to c2 and returns copy
                         // so either object can be "promotable"
                         // allowing unconstrained mixed mode
                         // arithmetic

  // indexing operator - must be a member function
    char& operator[](int n);
                         // x = c[3] = y;
                         // returning reference allows writing
                         // as well as reading indexed data

  // alternate indexing operator - must be a member function
    char& operator()(int n, int m);
                          // illustrates that double dimensions work

  // I/O operators - should be friend returning reference to stream
    friend ostream& operator<<(ostream& ostrm, C& c);
                         // cout << c1 << c2 << "\n";
                         // returning reference allows chaining

    friend istream& operator>>(istream& ostrm, C& c);
                         // cin >> c;

  // member function
    T& fun1(T& y2);       // y1 = c.fun1(y2);
                          // syntax is function attached to object

  // friend function
    friend T& fun2(C& c, T& y2);
                          // y1 = fun2(c, y2);
                          // syntax looks procedural and first
                          // argument can be an object promotable
                          // to a C

};                        // don't forget final semicolon
#endif C_HPP
